{% extends 'base.html' %}
{% load static %}


{% block content %}

    <div class="jumbotron jumbotron-fluid bg-primary text-center text-white">
        <h1>Geometric Metric Brownian Motion</h1>
        <p class="lead">Making Sense of the Noise</p>
    </div>

    <div class="container text-center">
        <div class="row">
            <div class="col-md-12">
                {{ script | safe }}
            </div>
        </div>
    </div>

{% endblock %}
